@@ -1,7 +1,7 @@
 #ifndef USERSPEC_H
 # define USERSPEC_H 1
 
-# include <stddef.h>
+# include <sys/types.h>
 
 const char *
 parse_user_spec (const char *spec_arg, uid_t *uid, gid_t *gid,
