@@ -1,5 +1,5 @@
 /* du -- summarize disk usage
-   Copyright (C) 1988-1991, 1995-2005 Free Software Foundation, Inc.
+   Copyright (C) 1988-1991, 1995-2006 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -31,7 +31,6 @@
 #include <assert.h>
 #include "system.h"
 #include "argmatch.h"
-#include "dirname.h" /* for strip_trailing_slashes */
 #include "error.h"
 #include "exclude.h"
 #include "fprintftime.h"
@@ -681,7 +680,11 @@ main (int argc, char **argv)
   struct Tokens tok;
 
   /* Bit flags that control how fts works.  */
-  int bit_flags = FTS_PHYSICAL | FTS_TIGHT_CYCLE_CHECK;
+  int bit_flags = FTS_TIGHT_CYCLE_CHECK;
+
+  /* Select one of the three FTS_ options that control if/when
+     to follow a symlink.  */
+  int symlink_deref_bits = FTS_PHYSICAL;
 
   /* If true, display only a total for each argument. */
   bool opt_summarize_only = false;
@@ -803,15 +806,15 @@ main (int argc, char **argv)
 	  break;
 
 	case 'D': /* This will eventually be 'H' (-H), too.  */
-	  bit_flags = FTS_COMFOLLOW;
+	  symlink_deref_bits = FTS_COMFOLLOW | FTS_PHYSICAL;
 	  break;
 
 	case 'L': /* --dereference */
-	  bit_flags = FTS_LOGICAL;
+	  symlink_deref_bits = FTS_LOGICAL;
 	  break;
 
 	case 'P': /* --no-dereference */
-	  bit_flags = FTS_PHYSICAL;
+	  symlink_deref_bits = FTS_PHYSICAL;
 	  break;
 
 	case 'S':
@@ -1000,6 +1003,7 @@ main (int argc, char **argv)
     ok = (i == j);
   }
 
+  bit_flags |= symlink_deref_bits;
   ok &= du_files (files, bit_flags);
 
   /* This isn't really necessary, but it does ensure we
