@@ -1,5 +1,5 @@
 /* GNU fmt -- simple text formatter.
-   Copyright (C) 1994-2003 Free Software Foundation, Inc.
+   Copyright (C) 1994-2006 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 /* Written by Ross Paterson <rap@doc.ic.ac.uk>.  */
 
@@ -56,7 +56,12 @@
    or too long.  The definition of SHORT_COST(n) should not be changed.
    However, EQUIV(n) may need tuning.  */
 
-typedef long COST;
+/* FIXME: "fmt" misbehaves given large inputs or options.  One
+   possible workaround for part of the problem is to change COST to be
+   a floating-point type.  There are other problems besides COST,
+   though; see MAXWORDS below.  */
+
+typedef long int COST;
 
 #define MAXCOST	TYPE_MAXIMUM (COST)
 
@@ -100,8 +105,12 @@ typedef long COST;
 #define LINE_CREDIT	EQUIV(3)
 
 /* Size of paragraph buffer, in words and characters.  Longer paragraphs
-   are handled neatly (cf. flush_paragraph()), so there's little to gain
-   by making these larger.  */
+   are handled neatly (cf. flush_paragraph()), so long as these values
+   are considerably greater than required by the width.  These values
+   cannot be extended indefinitely: doing so would run into size limits
+   and/or cause more overflows in cost calculations.  FIXME: Remove these
+   arbitrary limits.  */
+
 #define MAXWORDS	1000
 #define MAXCHARS	5000
 
@@ -115,13 +124,6 @@ typedef long COST;
    output.  */
 #define TABWIDTH	8
 
-/* Miscellaneous definitions.  */
-
-#undef TRUE
-#define TRUE true
-#undef FALSE
-#define FALSE false
-
 /* Word descriptor structure.  */
 
 typedef struct Word WORD;
@@ -171,16 +173,16 @@ const char *program_name;
 
 /* Option values.  */
 
-/* If TRUE, first 2 lines may have different indent (default FALSE).  */
+/* If true, first 2 lines may have different indent (default false).  */
 static bool crown;
 
-/* If TRUE, first 2 lines _must_ have different indent (default FALSE).  */
+/* If true, first 2 lines _must_ have different indent (default false).  */
 static bool tagged;
 
-/* If TRUE, each line is a paragraph on its own (default FALSE).  */
+/* If true, each line is a paragraph on its own (default false).  */
 static bool split;
 
-/* If TRUE, don't preserve inter-word spacing (default FALSE).  */
+/* If true, don't preserve inter-word spacing (default false).  */
 static bool uniform;
 
 /* Prefix minus leading and trailing spaces (default "").  */
@@ -228,7 +230,7 @@ static WORD word[MAXWORDS];
    word.  */
 static WORD *word_limit;
 
-/* If TRUE, current input file contains tab characters, and so tabs can be
+/* If true, current input file contains tab characters, and so tabs can be
    used for white space on output.  */
 static bool tabs;
 
@@ -263,7 +265,7 @@ static int last_line_length;
 void
 usage (int status)
 {
-  if (status != 0)
+  if (status != EXIT_SUCCESS)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
   else
@@ -279,7 +281,8 @@ Mandatory arguments to long options are mandatory for short options too.\n\
 "), stdout);
       fputs (_("\
   -c, --crown-margin        preserve indentation of first two lines\n\
-  -p, --prefix=STRING       combine only lines having STRING as prefix\n\
+  -p, --prefix=STRING       reformat only lines beginning with STRING,\n\
+                              reattaching the prefix to reformatted lines\n\
   -s, --split-only          split long lines, but do not refill\n\
 "),
 	     stdout);
@@ -296,7 +299,7 @@ With no FILE, or when FILE is -, read standard input.\n"),
 	     stdout);
       printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
-  exit (status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
+  exit (status);
 }
 
 /* Decode options and launch execution.  */
@@ -311,14 +314,15 @@ static const struct option long_options[] =
   {"width", required_argument, NULL, 'w'},
   {GETOPT_HELP_OPTION_DECL},
   {GETOPT_VERSION_OPTION_DECL},
-  {0, 0, 0, 0},
+  {NULL, 0, NULL, 0},
 };
 
 int
-main (register int argc, register char **argv)
+main (int argc, char **argv)
 {
   int optchar;
-  int fail;
+  bool ok = true;
+  char const *max_width_option = NULL;
 
   initialize_main (&argc, &argv);
   program_name = argv[0];
@@ -328,30 +332,17 @@ main (register int argc, register char **argv)
 
   atexit (close_stdout);
 
-  crown = tagged = split = uniform = FALSE;
+  crown = tagged = split = uniform = false;
   max_width = WIDTH;
   prefix = "";
   prefix_length = prefix_lead_space = prefix_full_length = 0;
 
   if (argc > 1 && argv[1][0] == '-' && ISDIGIT (argv[1][1]))
     {
-      const char *s = argv[1] + 1;
-      max_width = 0;
-      /* Old option syntax; a dash followed by one or more digits.
-         Move past the number. */
-      for (; ISDIGIT (*s); ++s)
-	{
-	  int old_max = max_width;
-	  max_width = max_width * 10 + *s - '0';
-	  if (INT_MAX / 10 < old_max || max_width < old_max)
-	    break;
-	}
+      /* Old option syntax; a dash followed by one or more digits.  */
+      max_width_option = argv[1] + 1;
 
-      if (*s)
-	error (EXIT_FAILURE, 0, _("invalid width option: %s"),
-	       quote (argv[1]));
-
-      /* Make the options we just parsed invisible to getopt. */
+      /* Make the option we just parsed invisible to getopt.  */
       argv[1] = argv[0];
       argv++;
       argc--;
@@ -370,29 +361,23 @@ main (register int argc, register char **argv)
 	usage (EXIT_FAILURE);
 
       case 'c':
-	crown = TRUE;
+	crown = true;
 	break;
 
       case 's':
-	split = TRUE;
+	split = true;
 	break;
 
       case 't':
-	tagged = TRUE;
+	tagged = true;
 	break;
 
       case 'u':
-	uniform = TRUE;
+	uniform = true;
 	break;
 
       case 'w':
-	{
-	  long int tmp_long;
-	  if (xstrtol (optarg, NULL, 10, &tmp_long, "") != LONGINT_OK
-	      || tmp_long <= 0 || tmp_long > INT_MAX)
-	    error (EXIT_FAILURE, 0, _("invalid width: %s"), quote (optarg));
-	  max_width = (int) tmp_long;
-	}
+	max_width_option = optarg;
 	break;
 
       case 'p':
@@ -405,9 +390,20 @@ main (register int argc, register char **argv)
 
       }
 
+  if (max_width_option)
+    {
+      /* Limit max_width to MAXCHARS / 2; otherwise, the resulting
+	 output can be quite ugly.  */
+      unsigned long int tmp;
+      if (! (xstrtoul (max_width_option, NULL, 10, &tmp, "") == LONGINT_OK
+	     && tmp <= MAXCHARS / 2))
+	error (EXIT_FAILURE, 0, _("invalid width: %s"),
+	       quote (max_width_option));
+      max_width = tmp;
+    }
+
   best_width = max_width * (2 * (100 - LEEWAY) + 1) / 200;
 
-  fail = 0;
   if (optind == argc)
     fmt (stdin);
   else
@@ -427,29 +423,29 @@ main (register int argc, register char **argv)
 		  if (fclose (in_stream) == EOF)
 		    {
 		      error (0, errno, "%s", file);
-		      fail = 1;
+		      ok = false;
 		    }
 		}
 	      else
 		{
 		  error (0, errno, _("cannot open %s for reading"),
 			 quote (file));
-		  fail = 1;
+		  ok = false;
 		}
 	    }
 	}
     }
 
-  exit (fail ? EXIT_FAILURE : EXIT_SUCCESS);
+  exit (ok ? EXIT_SUCCESS : EXIT_FAILURE);
 }
 
 /* Trim space from the front and back of the string P, yielding the prefix,
    and record the lengths of the prefix and the space trimmed.  */
 
 static void
-set_prefix (register char *p)
+set_prefix (char *p)
 {
-  register char *s;
+  char *s;
 
   prefix_lead_space = 0;
   while (*p == ' ')
@@ -471,7 +467,7 @@ set_prefix (register char *p)
 static void
 fmt (FILE *f)
 {
-  tabs = FALSE;
+  tabs = false;
   other_indent = 0;
   next_char = get_prefix (f);
   while (get_paragraph (f))
@@ -526,13 +522,13 @@ set_other_indent (bool same_paragraph)
    If a prefix is in effect, it must be present at the same indent for
    each line in the paragraph.
 
-   Return FALSE if end-of-file was encountered before the start of a
-   paragraph, else TRUE.  */
+   Return false if end-of-file was encountered before the start of a
+   paragraph, else true.  */
 
 static bool
 get_paragraph (FILE *f)
 {
-  register int c;
+  int c;
 
   last_line_length = 0;
   c = next_char;
@@ -547,7 +543,7 @@ get_paragraph (FILE *f)
       if (c == EOF)
 	{
 	  next_char = EOF;
-	  return FALSE;
+	  return false;
 	}
       putchar ('\n');
       c = get_prefix (f);
@@ -595,9 +591,9 @@ get_paragraph (FILE *f)
       while (same_para (c) && in_column == other_indent)
 	c = get_line (f, c);
     }
-  (word_limit - 1)->period = (word_limit - 1)->final = TRUE;
+  (word_limit - 1)->period = (word_limit - 1)->final = true;
   next_char = c;
-  return TRUE;
+  return true;
 }
 
 /* Copy to the output a line that failed to match the prefix, or that
@@ -606,16 +602,16 @@ get_paragraph (FILE *f)
    Return the character (\n or EOF) ending the line.  */
 
 static int
-copy_rest (FILE *f, register int c)
+copy_rest (FILE *f, int c)
 {
-  register const char *s;
+  const char *s;
 
   out_column = 0;
   if (in_column > next_prefix_indent && c != '\n' && c != EOF)
     {
       put_space (next_prefix_indent);
       for (s = prefix; out_column != in_column && *s; out_column++)
-	putchar (*(unsigned char *)s++);
+	putchar (*s++);
       put_space (in_column - out_column);
     }
   while (c != '\n' && c != EOF)
@@ -626,12 +622,12 @@ copy_rest (FILE *f, register int c)
   return c;
 }
 
-/* Return TRUE if a line whose first non-blank character after the
+/* Return true if a line whose first non-blank character after the
    prefix (if any) is C could belong to the current paragraph,
-   otherwise FALSE.  */
+   otherwise false.  */
 
 static bool
-same_para (register int c)
+same_para (int c)
 {
   return (next_prefix_indent == prefix_indent
 	  && in_column >= next_prefix_indent + prefix_full_length
@@ -647,11 +643,11 @@ same_para (register int c)
    Return the first non-blank character of the next line.  */
 
 static int
-get_line (FILE *f, register int c)
+get_line (FILE *f, int c)
 {
   int start;
-  register char *end_of_parabuf;
-  register WORD *end_of_word;
+  char *end_of_parabuf;
+  WORD *end_of_word;
 
   end_of_parabuf = &parabuf[MAXCHARS];
   end_of_word = &word[MAXWORDS - 2];
@@ -672,7 +668,7 @@ get_line (FILE *f, register int c)
 	  *wptr++ = c;
 	  c = getc (f);
 	}
-      while (c != EOF && !ISSPACE (c));
+      while (c != EOF && !isspace (c));
       in_column += word_limit->length = wptr - word_limit->text;
       check_punctuation (word_limit);
 
@@ -705,7 +701,7 @@ get_line (FILE *f, register int c)
 static int
 get_prefix (FILE *f)
 {
-  register int c;
+  int c;
 
   in_column = 0;
   c = get_space (f, getc (f));
@@ -718,7 +714,8 @@ get_prefix (FILE *f)
       next_prefix_indent = in_column;
       for (p = prefix; *p != '\0'; p++)
 	{
-	  if (c != *(unsigned char *)p)
+	  unsigned char pc = *p;
+	  if (c != pc)
 	    return c;
 	  in_column++;
 	  c = getc (f);
@@ -732,7 +729,7 @@ get_prefix (FILE *f)
    in_column up-to-date.  Return first non-blank character.  */
 
 static int
-get_space (FILE *f, register int c)
+get_space (FILE *f, int c)
 {
   for (;;)
     {
@@ -740,7 +737,7 @@ get_space (FILE *f, register int c)
 	in_column++;
       else if (c == '\t')
 	{
-	  tabs = TRUE;
+	  tabs = true;
 	  in_column = (in_column / TABWIDTH + 1) * TABWIDTH;
 	}
       else
@@ -752,15 +749,15 @@ get_space (FILE *f, register int c)
 /* Set extra fields in word W describing any attached punctuation.  */
 
 static void
-check_punctuation (register WORD *w)
+check_punctuation (WORD *w)
 {
-  const unsigned char *start, *finish;
+  char const *start = w->text;
+  char const *finish = start + (w->length - 1);
+  unsigned char fin = *finish;
 
-  start = (unsigned char *) w->text;
-  finish = start + (w->length - 1);
   w->paren = isopen (*start);
-  w->punct = ISPUNCT (*finish);
-  while (isclose (*finish) && finish > start)
+  w->punct = !! ispunct (fin);
+  while (start < finish && isclose (*finish))
     finish--;
   w->period = isperiod (*finish);
 }
@@ -772,7 +769,7 @@ static void
 flush_paragraph (void)
 {
   WORD *split_point;
-  register WORD *w;
+  WORD *w;
   int shift;
   COST best_break;
 
@@ -780,7 +777,7 @@ flush_paragraph (void)
 
   if (word_limit == word)
     {
-      printf ("%*s", (int) (wptr - parabuf), parabuf);
+      fwrite (parabuf, sizeof *parabuf, wptr - parabuf, stdout);
       wptr = parabuf;
       return;
     }
@@ -812,7 +809,7 @@ flush_paragraph (void)
   /* Copy text of words down to start of parabuf -- we use memmove because
      the source and target may overlap.  */
 
-  memmove (parabuf, split_point->text, (size_t) (wptr - split_point->text));
+  memmove (parabuf, split_point->text, wptr - split_point->text);
   shift = split_point->text - parabuf;
   wptr -= shift;
 
@@ -824,8 +821,7 @@ flush_paragraph (void)
   /* Copy words from split_point down to word -- we use memmove because
      the source and target may overlap.  */
 
-  memmove ((char *) word, (char *) split_point,
-	 (word_limit - split_point + 1) * sizeof (WORD));
+  memmove (word, split_point, (word_limit - split_point + 1) * sizeof *word);
   word_limit -= split_point - word;
 }
 
@@ -836,9 +832,9 @@ flush_paragraph (void)
 static void
 fmt_paragraph (void)
 {
-  register WORD *start, *w;
-  register int len;
-  register COST wcost, best;
+  WORD *start, *w;
+  int len;
+  COST wcost, best;
   int saved_length;
 
   word_limit->best_cost = 0;
@@ -890,9 +886,9 @@ fmt_paragraph (void)
    word THIS.  */
 
 static COST
-base_cost (register WORD *this)
+base_cost (WORD *this)
 {
-  register COST cost;
+  COST cost;
 
   cost = LINE_COST;
 
@@ -923,10 +919,10 @@ base_cost (register WORD *this)
    depends on LEN, the length of the line beginning there.  */
 
 static COST
-line_cost (register WORD *next, register int len)
+line_cost (WORD *next, int len)
 {
-  register int n;
-  register COST cost;
+  int n;
+  COST cost;
 
   if (next == word_limit)
     return 0;
@@ -944,9 +940,9 @@ line_cost (register WORD *next, register int len)
    FINISH, which must be in the next_break chain from word.  */
 
 static void
-put_paragraph (register WORD *finish)
+put_paragraph (WORD *finish)
 {
-  register WORD *w;
+  WORD *w;
 
   put_line (word, first_indent);
   for (w = word->next_break; w != finish; w = w->next_break)
@@ -957,9 +953,9 @@ put_paragraph (register WORD *finish)
    INDENT, including the prefix (if any).  */
 
 static void
-put_line (register WORD *w, int indent)
+put_line (WORD *w, int indent)
 {
-  register WORD *endline;
+  WORD *endline;
 
   out_column = 0;
   put_space (prefix_indent);
@@ -981,10 +977,10 @@ put_line (register WORD *w, int indent)
 /* Output to stdout the word W.  */
 
 static void
-put_word (register WORD *w)
+put_word (WORD *w)
 {
-  register const char *s;
-  register int n;
+  const char *s;
+  int n;
 
   s = w->text;
   for (n = w->length; n != 0; n--)
@@ -997,7 +993,7 @@ put_word (register WORD *w)
 static void
 put_space (int space)
 {
-  register int space_target, tab_target;
+  int space_target, tab_target;
 
   space_target = out_column + space;
   if (tabs)
