@@ -1,5 +1,5 @@
 /* tac - concatenate and print files in reverse
-   Copyright (C) 1988-1991, 1995-2005 Free Software Foundation, Inc.
+   Copyright (C) 1988-1991, 1995-2006 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -78,7 +78,7 @@ tac -r -s '.\|
 char *program_name;
 
 /* The string that separates the records of the file. */
-static char *separator;
+static char const *separator;
 
 /* True if we have ever read standard input.  */
 static bool have_read_stdin = false;
@@ -110,6 +110,7 @@ static size_t G_buffer_size;
 
 /* The compiled regular expression representing `separator'. */
 static struct re_pattern_buffer compiled_separator;
+static char compiled_separator_fastmap[UCHAR_MAX + 1];
 
 static struct option const longopts[] =
 {
@@ -210,7 +211,7 @@ tac_seekable (int input_fd, const char *file)
      Only used when the separator is attached to the preceding record. */
   bool first_time = true;
   char first_char = *separator;	/* Speed optimization, non-regexp. */
-  char *separator1 = separator + 1; /* Speed optimization, non-regexp. */
+  char const *separator1 = separator + 1; /* Speed optimization, non-regexp. */
   size_t match_length1 = match_length - 1; /* Speed optimization, non-regexp. */
   struct re_registers regs;
 
@@ -425,7 +426,7 @@ static bool
 copy_to_temp (FILE **g_tmp, char **g_tempfile, int input_fd, char const *file)
 {
   static char *template = NULL;
-  static char *tempdir;
+  static char const *tempdir;
   char *tempfile;
   FILE *tmp;
   int fd;
@@ -608,9 +609,9 @@ main (int argc, char **argv)
 
   if (sentinel_length == 0)
     {
-      compiled_separator.allocated = 100;
-      compiled_separator.buffer = xmalloc (compiled_separator.allocated);
-      compiled_separator.fastmap = xmalloc (256);
+      compiled_separator.buffer = NULL;
+      compiled_separator.allocated = 0;
+      compiled_separator.fastmap = compiled_separator_fastmap;
       compiled_separator.translate = NULL;
       error_message = re_compile_pattern (separator, strlen (separator),
 					  &compiled_separator);
